import crypto from 'crypto';
import * as cborg from 'cborg';
import { INTEGRITY_BLOCK_MAGIC, VERSION_B1, VERSION_B2, } from '../utils/constants.js';
import { checkDeterministic } from '../cbor/deterministic.js';
import { getRawPublicKey, checkIsValidKey, getPublicKeyAttributeName, } from '../utils/utils.js';
export class IntegrityBlockSigner {
    webBundle;
    webBundleId;
    signingStrategies;
    is_v2;
    // `webBundleId` is ignored if `is_v2` is false.
    constructor(webBundle, webBundleId, signingStrategies, is_v2 = true) {
        this.webBundle = webBundle;
        this.webBundleId = webBundleId;
        this.signingStrategies = signingStrategies;
        this.is_v2 = is_v2;
    }
    async sign() {
        const integrityBlock = this.obtainIntegrityBlock().integrityBlock;
        if (this.is_v2) {
            integrityBlock.setWebBundleId(this.webBundleId);
        }
        const signatures = new Array();
        for (const signingStrategy of this.signingStrategies) {
            const publicKey = await signingStrategy.getPublicKey();
            checkIsValidKey('public', publicKey);
            const newAttributes = {
                [getPublicKeyAttributeName(publicKey)]: getRawPublicKey(publicKey),
            };
            const ibCbor = integrityBlock.toCBOR();
            const attrCbor = cborg.encode(newAttributes);
            checkDeterministic(ibCbor);
            checkDeterministic(attrCbor);
            const dataToBeSigned = this.generateDataToBeSigned(this.calcWebBundleHash(), ibCbor, attrCbor);
            const signature = await signingStrategy.sign(dataToBeSigned);
            this.verifySignature(dataToBeSigned, signature, publicKey);
            // The signatures are calculated independently, and thus not appended to
            // the integrity block here to not affect subsequent calculations.
            signatures.push({
                signature,
                signatureAttributes: newAttributes,
            });
        }
        for (const signature of signatures) {
            integrityBlock.addIntegritySignature(signature);
        }
        const signedIbCbor = integrityBlock.toCBOR();
        checkDeterministic(signedIbCbor);
        return {
            integrityBlock: signedIbCbor,
            signedWebBundle: new Uint8Array(Buffer.concat([signedIbCbor, this.webBundle])),
        };
    }
    readWebBundleLength() {
        // The length of the web bundle is contained in the last 8 bytes of the web
        // bundle, represented as BigEndian.
        let buffer = Buffer.from(this.webBundle.slice(-8));
        // Number is big enough to represent 4GB which is the limit for the web
        // bundle size which can be contained in a Buffer, which is the format
        // in which it's passed back to e.g. Webpack.
        return Number(buffer.readBigUint64BE());
    }
    obtainIntegrityBlock() {
        const webBundleLength = this.readWebBundleLength();
        if (webBundleLength !== this.webBundle.length) {
            throw new Error('IntegrityBlockSigner: Re-signing signed bundles is not supported yet.');
        }
        return { integrityBlock: new IntegrityBlock(this.is_v2), offset: 0 };
    }
    calcWebBundleHash() {
        var hash = crypto.createHash('sha512');
        var data = hash.update(this.webBundle);
        return new Uint8Array(data.digest());
    }
    generateDataToBeSigned(webBundleHash, integrityBlockCborBytes, newAttributesCborBytes) {
        // The order is critical and must be the following:
        // (0) hash of the bundle,
        // (1) integrity block, and
        // (2) attributes.
        const dataParts = [
            webBundleHash,
            integrityBlockCborBytes,
            newAttributesCborBytes,
        ];
        const bigEndianNumLength = 8;
        const totalLength = dataParts.reduce((previous, current) => {
            return previous + current.length;
        }, /*one big endian num per part*/ dataParts.length * bigEndianNumLength);
        let buffer = Buffer.alloc(totalLength);
        let offset = 0;
        dataParts.forEach((d) => {
            buffer.writeBigInt64BE(BigInt(d.length), offset);
            offset += bigEndianNumLength;
            Buffer.from(d).copy(buffer, offset);
            offset += d.length;
        });
        return new Uint8Array(buffer);
    }
    verifySignature(data, signature, publicKey) {
        // For ECDSA P-256 keys the algorithm is implicitly selected as SHA-256.
        const isVerified = crypto.verify(
        /*algorithm=*/ undefined, data, publicKey, signature);
        if (!isVerified) {
            throw new Error('IntegrityBlockSigner: Signature cannot be verified. Your keys might be corrupted or not corresponding each other.');
        }
    }
}
export class IntegrityBlock {
    is_v2;
    attributes = new Map();
    signatureStack = [];
    constructor(is_v2 = false) {
        this.is_v2 = is_v2;
    }
    setWebBundleId(webBundleId) {
        if (!this.is_v2) {
            throw new Error('setWebBundleId() is only available for v2 integrity blocks.');
        }
        this.attributes.set('webBundleId', webBundleId);
    }
    addIntegritySignature(is) {
        this.signatureStack.push(is);
    }
    toCBOR() {
        if (this.is_v2) {
            return cborg.encode([
                INTEGRITY_BLOCK_MAGIC,
                VERSION_B2,
                this.attributes,
                this.signatureStack.map((integritySig) => {
                    // The CBOR must have an array of length 2 containing the following:
                    // (0) attributes and (1) signature. The order is important.
                    return [integritySig.signatureAttributes, integritySig.signature];
                }),
            ]);
        }
        else {
            return cborg.encode([
                INTEGRITY_BLOCK_MAGIC,
                VERSION_B1,
                this.signatureStack.map((integritySig) => {
                    // The CBOR must have an array of length 2 containing the following:
                    // (0) attributes and (1) signature. The order is important.
                    return [integritySig.signatureAttributes, integritySig.signature];
                }),
            ]);
        }
    }
}
