import crypto from 'crypto';
import base32Encode from 'base32-encode';
import { getRawPublicKey, isAsymmetricKeyTypeSupported, getSignatureType, } from './utils/utils.js';
import { SignatureType } from './utils/constants.js';
// Web Bundle ID is a base32-encoded (without padding) ed25519 public key
// transformed to lowercase. More information:
// https://github.com/WICG/isolated-web-apps/blob/main/Scheme.md#signed-web-bundle-ids
export class WebBundleId {
    // https://github.com/WICG/isolated-web-apps/blob/main/Scheme.md#suffix
    TYPE_SUFFIX_MAPPING = new Map([
        [SignatureType.Ed25519, [0x00, 0x01, 0x02]],
        [SignatureType.EcdsaP256SHA256, [0x00, 0x02, 0x02]],
    ]);
    scheme = 'isolated-app://';
    key;
    typeSuffix;
    constructor(key) {
        if (!isAsymmetricKeyTypeSupported(key)) {
            throw new Error(`WebBundleId: Only Ed25519 and ECDSA P-256 keys are currently supported.`);
        }
        if (key.type === 'private') {
            this.key = crypto.createPublicKey(key);
        }
        else {
            this.key = key;
        }
        this.typeSuffix = this.TYPE_SUFFIX_MAPPING.get(getSignatureType(this.key));
    }
    serialize() {
        return base32Encode(new Uint8Array([...getRawPublicKey(this.key), ...this.typeSuffix]), 'RFC4648', { padding: false }).toLowerCase();
    }
    serializeWithIsolatedWebAppOrigin() {
        return `${this.scheme}${this.serialize()}/`;
    }
    toString() {
        return `\
  Web Bundle ID: ${this.serialize()}
  Isolated Web App Origin: ${this.serializeWithIsolatedWebAppOrigin()}`;
    }
}
